summaryrefslogtreecommitdiff
path: root/drivers/gpu/drm/drm_bridge.c
diff options
context:
space:
mode:
Diffstat (limited to 'drivers/gpu/drm/drm_bridge.c')
-rw-r--r--drivers/gpu/drm/drm_bridge.c134
1 files changed, 119 insertions, 15 deletions
diff --git a/drivers/gpu/drm/drm_bridge.c b/drivers/gpu/drm/drm_bridge.c
index 8f355df883d8..3b165a0d1e77 100644
--- a/drivers/gpu/drm/drm_bridge.c
+++ b/drivers/gpu/drm/drm_bridge.c
@@ -27,6 +27,7 @@
#include <linux/media-bus-format.h>
#include <linux/module.h>
#include <linux/mutex.h>
+#include <linux/srcu.h>
#include <drm/drm_atomic_state_helper.h>
#include <drm/drm_bridge.h>
@@ -202,6 +203,67 @@ static DEFINE_MUTEX(bridge_lock);
static LIST_HEAD(bridge_list);
static LIST_HEAD(bridge_lingering_list);
+DEFINE_STATIC_SRCU(drm_bridge_unplug_srcu);
+
+/**
+ * drm_bridge_enter - Enter DRM bridge critical section
+ * @bridge: DRM bridge
+ * @idx: Pointer to index that will be passed to the matching drm_bridge_exit()
+ *
+ * This function marks and protects the beginning of a section that should not
+ * be entered after the bridge has been unplugged. The section end is marked
+ * with drm_bridge_exit(). Calls to this function can be nested.
+ *
+ * Returns:
+ * True if it is OK to enter the section, false otherwise.
+ */
+bool drm_bridge_enter(struct drm_bridge *bridge, int *idx)
+{
+ *idx = srcu_read_lock(&drm_bridge_unplug_srcu);
+
+ if (bridge->unplugged) {
+ srcu_read_unlock(&drm_bridge_unplug_srcu, *idx);
+ return false;
+ }
+
+ return true;
+}
+EXPORT_SYMBOL(drm_bridge_enter);
+
+/**
+ * drm_bridge_exit - Exit DRM bridge critical section
+ * @idx: index returned by drm_bridge_enter()
+ *
+ * This function marks the end of a section that should not be entered after
+ * the bridge has been unplugged.
+ */
+void drm_bridge_exit(int idx)
+{
+ srcu_read_unlock(&drm_bridge_unplug_srcu, idx);
+}
+EXPORT_SYMBOL(drm_bridge_exit);
+
+/**
+ * drm_bridge_unplug - declare a DRM bridge was unplugged and remove it
+ * @bridge: DRM bridge
+ *
+ * This tells the bridge has been physically unplugged and no operations on
+ * device resources must be done anymore. Entry-points can use
+ * drm_bridge_enter() and drm_bridge_exit() to protect device resources in
+ * a race free manner.
+ *
+ * Also unregisters the bridge.
+ */
+void drm_bridge_unplug(struct drm_bridge *bridge)
+{
+ bridge->unplugged = true;
+
+ synchronize_srcu(&drm_bridge_unplug_srcu);
+
+ drm_bridge_remove(bridge);
+}
+EXPORT_SYMBOL(drm_bridge_unplug);
+
static void __drm_bridge_free(struct kref *kref)
{
struct drm_bridge *bridge = container_of(kref, struct drm_bridge, refcount);
@@ -213,6 +275,8 @@ static void __drm_bridge_free(struct kref *kref)
if (bridge->funcs->destroy)
bridge->funcs->destroy(bridge);
+ drm_bridge_put(bridge->next_bridge);
+
kfree(bridge->container);
}
@@ -299,7 +363,7 @@ EXPORT_SYMBOL(__devm_drm_bridge_alloc);
* @bridge: bridge control structure
*
* Add the given bridge to the global list of bridges, where they can be
* found by users via of_drm_find_bridge().
*
* The bridge to be added must have been allocated by
* devm_drm_bridge_alloc().
@@ -360,9 +424,9 @@ EXPORT_SYMBOL(devm_drm_bridge_add);
* @bridge: bridge control structure
*
* Remove the given bridge from the global list of registered bridges, so
* it won't be found by users via of_drm_find_bridge(), and add it to the
* lingering bridge list, to keep track of it until its allocated memory is
* eventually freed.
*/
void drm_bridge_remove(struct drm_bridge *bridge)
{
@@ -1418,29 +1482,69 @@ EXPORT_SYMBOL_GPL(drm_bridge_hpd_notify);
#ifdef CONFIG_OF
/**
+ * of_drm_find_and_get_bridge - find the bridge corresponding to the device
+ * node in the global bridge list
+ * @np: device node
+ *
+ * The refcount of the returned bridge is incremented. Use drm_bridge_put()
+ * when done with it.
+ *
+ * RETURNS:
+ * drm_bridge control struct on success, NULL on failure
+ */
+struct drm_bridge *of_drm_find_and_get_bridge(struct device_node *np)
+{
+ struct drm_bridge *bridge;
+
+ scoped_guard(mutex, &bridge_lock) {
+ list_for_each_entry(bridge, &bridge_list, list)
+ if (bridge->of_node == np)
+ return drm_bridge_get(bridge);
+ }
+
+ return NULL;
+}
+EXPORT_SYMBOL(of_drm_find_and_get_bridge);
+
+/**
* of_drm_find_bridge - find the bridge corresponding to the device node in
* the global bridge list
*
* @np: device node
*
+ * This function is deprecated. Convert to of_drm_find_and_get_bridge()
+ * instead for proper refcounting.
+ *
+ * The bridge returned by this function is not refcounted. This is
+ * dangerous because the bridge might be deallocated even before the caller
+ * has a chance to use it. To use this function you have to do one of:
+ *
+ * - get a reference with drm_bridge_get() as soon as possible to
+ * minimize the race window, and then drm_bridge_put() when no longer
+ * using the pointer
+ *
+ * - not call drm_bridge_get() or drm_bridge_put() at all, which used to
+ * be the correct practice before dynamic bridge lifetime was introduced
+ *
+ * - again, convert to of_drm_find_and_get_bridge(), which is the only safe
+ * thing to do
+ *
* RETURNS:
* drm_bridge control struct on success, NULL on failure
*/
struct drm_bridge *of_drm_find_bridge(struct device_node *np)
{
- struct drm_bridge *bridge;
-
- mutex_lock(&bridge_lock);
+ struct drm_bridge *bridge = of_drm_find_and_get_bridge(np);
- list_for_each_entry(bridge, &bridge_list, list) {
- if (bridge->of_node == np) {
- mutex_unlock(&bridge_lock);
- return bridge;
- }
- }
+ /*
+ * We need to emulate the original semantics of
+ * of_drm_find_bridge(), which was not getting any bridge
+ * reference. Being now based on of_drm_find_and_get_bridge() which
+ * gets a reference, put it before returning.
+ */
+ drm_bridge_put(bridge);
- mutex_unlock(&bridge_lock);
- return NULL;
+ return bridge;
}
EXPORT_SYMBOL(of_drm_find_bridge);
#endif
@@ -299,7 +363,7 @@ EXPORT_SYMBOL(__devm_drm_bridge_alloc);
*
* Add the given bridge to the global list of bridges, where they can be
- * found by users via of_drm_find_bridge().
+ * found by users via of_drm_find_and_get_bridge().
*
* The bridge to be added must have been allocated by
@@ -360,9 +424,9 @@ EXPORT_SYMBOL(devm_drm_bridge_add);
*
* Remove the given bridge from the global list of registered bridges, so
- * it won't be found by users via of_drm_find_bridge(), and add it to the
- * lingering bridge list, to keep track of it until its allocated memory is
- * eventually freed.
+ * it won't be found by users via of_drm_find_and_get_bridge(), and add it
+ * to the lingering bridge list, to keep track of it until its allocated
+ * memory is eventually freed.
*/
void drm_bridge_remove(struct drm_bridge *bridge)